/*

The NetUsage Project
http://netusage.mozdev.org/

Copyright (C) 2006, NetUsage Project
This extension can be freely distributed under
the terms of the Mozilla Public Licence 1.1
See the licence at http://www.mozilla.org/MPL/

---------
Net Usage Item
XPCOM

*/

const nui_contract_id = "@mozilla.org/netusageitem;1";
const nui_class_id = "{489e3485-7a4a-4438-9185-bf14c0c317ba}";



nsNetUsageItem.prototype = {

	//
	// nsINetUsageItem
	//

	get itemid() { return this.mitemid; },
	get status() { return this.mStatus; },
	get errorText() { return this.mErrorText; },
	get lastFetched() { return this.mLastFetched; },
	get ispid() { return this.isp ? this.isp.id : ''; },
	getVar: function(k) { return this.vars[k]; },

	//
	// private
	//

	init: function(aitemid) {
		// includes
		var scriptinc = Components.classes["@mozilla.org/moz/jssubscript-loader;1"].getService().QueryInterface(Components.interfaces.mozIJSSubScriptLoader);
		
		if (typeof(__USAGELIB_JS__) == 'undefined')
			scriptinc.loadSubScript("chrome://netusageitem/content/usagelib.js");
		this.getISPInterface = nuGetISPInterface;

		this.observerSvc = Components.classes["@mozilla.org/observer-service;1"].getService().QueryInterface(Components.interfaces.nsIObserverService);

		ndump('XXXX init: '+aitemid);

		this.mitemid = aitemid;
		this.checkInterval = -1;
		this.isp = null;
		this.mLastFetched = 0;
		
		this.timer = Components.classes['@mozilla.org/timer;1'].createInstance(Components.interfaces.nsITimer);

		var prefSvc = Components.classes["@mozilla.org/preferences-service;1"].getService(Components.interfaces.nsIPrefService);
		//var pref = prefSvc.getBranch("extensions.netusage."+this.mitemid+".");
		var pref = prefSvc.getBranch(""); // TODO why doesn't above work?
		pref.QueryInterface(Components.interfaces.nsIPrefBranch2);
		pref.addObserver("extensions.netusage."+this.mitemid+".", this, false);
		ndump("extensions.netusage."+this.mitemid+".");
		
		
		// load prefs should trigger initTimer()
		this.loadPrefs();
	},
	
	uninit: function() {
		var prefSvc = Components.classes["@mozilla.org/preferences-service;1"].getService(Components.interfaces.nsIPrefService);
		var pref = prefSvc.getBranch("");
		pref.QueryInterface(Components.interfaces.nsIPrefBranch2);
		pref.removeObserver("extensions.netusage."+this.mitemid+".", this);
		this.timer.cancel();
		this.abort();
		//this.observerSvc.removeObserver(this, '');
		this.observerSvc = null;
		this.timer = null;
		ndump('UnInit !!!!!!!!!!!!!!!!');
	},

	// initialises timer
	initTimer: function () {
		this.timer.cancel();
		if (this.isp != null && this.checkInterval >= 600)
			this.timer.init(this, this.checkInterval * 1000, this.timer.TYPE_REPEATING_SLACK);
		else
			ndump('timer disabled');
	},

	//
	// nsINetUsageItem
	//
	
	setISPInterface: function(objjson) {
		this.isp = eval('(' + objjson + ')');
	},
	
	getVarsObjStr: function() {
		return ndumpobjstr(this.vars);
	},
	
	getVarExpand: function(s) {
		return nuVarExpand(s, this.vars, false, false, false, this.isp, 
			this.filterFuncList, this.filterFunc);
	},
	
	abort: function() {
		if (this.mStatus == 'fetching') {
			// abort
			ndump('aborting...');
			this.xmlhttp.abort();
			this.endUpdate('aborted', 'User aborted usage update');
		}
	},
	
	endUpdate: function(status, errorText) {
		ndump('Fetch ended');
			
		if (status != null)
			this.mStatus = status;
		if (errorText != null)
			this.mErrorText = errorText;
		
		// write or delete debug file?
		if (this.mDebug || this.status == 'error-process' || this.status == 'error-parse') {
			this.debugDump(this.mErrorText != '' ? this.mErrorText+"\n" : "Debug mode\n");
		} else {
			var file = Components.classes["@mozilla.org/file/directory_service;1"].getService(Components.interfaces.nsIProperties).get("ProfD", Components.interfaces.nsIFile);
			file.append("netusage_debug.txt");
			if (file.exists())
				file.remove(false);
		}
		
		try {
		this.observerSvc.removeObserver(this, 'http-on-examine-response');
		} catch(e) {}
		if (this.mOpenPageCmd == null)
			this.observerSvc.notifyObservers(this, 'netusage-update-gui', 'draw');
		else
			this.observerSvc.notifyObservers(this, 'netusage-open-page', this.mOpenPageCmd);

		this.xmlhttp = null;
		this.mDebughttp = '';
	},

	updateUsage: function(openPage) {
		try {

		if (this.isp == null) {
			ndump('updateUsage this.isp == null');
			this.mErrorText = 'Please select an ISP in your preferences.'
			return false;
		}
		
		if (this.mStatus == 'fetching') {
			// we're already fetching!
			return false;
		}
	
		this.mDebughttp = '';
		this.mStatus = 'fetching';
		// beyond this point, we will always return true and we will always notify to show error //
		
		// update all client displays (so they flash)
		this.observerSvc.notifyObservers(this, 'netusage-update-gui', 'draw');

		// clean step, vars
		this.step = 0;
		this.nextstep = 0;
		this.mVarMapped = false;
		this.mErrorText = "";
		this.mOpenPageCmd = null;
		this.mUsagePageStep = -1;
		if (openPage)
			this.prepareUsagePageOpen();
		
		this.vars = new Array();
		this.vars['ispname'] = this.isp.name;
		if (this.isp.vars) {
			// copy over isp vars
			for (i in this.isp.vars)
				this.vars[i] = this.isp.vars[i];
		}

		// fetch username and password
		var ublogin;
		if (!(ublogin = nuLoadPass('net:usage:bar:'+this.mitemid))) {
			this.endUpdate('error-up', "Username and Password required");
			return true;
		}
		this.vars['USERNAME'] = ublogin[1];
		this.vars['PASSWORD'] = ublogin[2];

		// TODO set this here or when we finish?
		this.mLastFetched = Math.floor((new Date()).getTime()/1000);
		
		// load [override] vars from pref
		nuLoadPrefList("extensions.netusage."+this.mitemid+".vars.", this.vars);
		
		// some predefined vars
		var now = new Date();
		var monthto3l = ['Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun', 'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec'];
		this.vars['now_month'] = now.getMonth()+1;
		this.vars['now_month_3l'] = monthto3l[now.getMonth()];
		this.vars['now_day'] = now.getDate();
		this.vars['now_year'] = now.getFullYear();
		
		// debug
		// ndumpobj('this.vars', this.vars);

		if (!('strToDate' in this.isp)) {
			this.isp.strToDate = function(m, s, isp) {
				var month3l = {'jan': 0, 'feb': 1, 'mar': 2, 'apr': 3, 'may': 4, 'jun': 5, 'jul': 6, 'aug': 7, 'sep': 8, 'oct': 9, 'nov': 10, 'dec': 11,
					'january': 0, 'february': 1, 'march': 2, 'april': 3, 'may': 4, 'june': 5, 'july': 6, 'august': 7, 'september': 8, 'october': 9, 'november': 10, 'december': 11};
				
				s = ''+s; // convert s to string (for s.split)
				if (m == 'roll' && s.length <= 2) {
					// roll day - lets make it a full roll date
					var now = new Date();
					var d = parseInt(s, 10);
					var mn = (d < now.getDate()) ? now.getMonth()+1+1 : now.getMonth()+1;
					s = d + '/' + mn + '/' + now.getFullYear();
				}

				// generic 25/10/2004 or 2004/10/25 or 2004-10-25
				var xpdtar, xtmp;
				xpdtar = s.split(/[-\/ ,]+/);
				if (parseInt(xpdtar[0], 10) > 1000) {
					// we have 2004/10/25: swap first and last date element
					xtmp = xpdtar[0];
					xpdtar[0] = xpdtar[2];
					xpdtar[2] = xtmp;
				}
				// check if 3+ letter months ie Jan/Feb/etc (27 May 2006 or 3 February, 2006)
				if (xpdtar[1].length >= 3)
					xpdtar[1] = month3l[xpdtar[1].toLowerCase()];
				else // UGLY F*(&(* JS Date has Months starting with 0
					xpdtar[1] = parseInt(xpdtar[1], 10)-1;
				ndump('strToDate: '+xpdtar[0]+'::'+xpdtar[1]+'+1::'+xpdtar[2]);

				if (m == 'start' || m == 'roll') {
					return new Date(parseInt(xpdtar[2], 10), parseInt(xpdtar[1], 10), parseInt(xpdtar[0], 10), 0, 0, 0);
				}
				else if (m == 'end') {
					return new Date(parseInt(xpdtar[2], 10), parseInt(xpdtar[1], 10), parseInt(xpdtar[0], 10), 23, 59, 59);
				}
			}
		}

		// for nuking WWW-authenticate
		this.observerSvc.addObserver(this, 'http-on-examine-response', false);
		
		// start the process
		this.resumeProcess();
		
		} catch(e) {
			this.endUpdate('error', 'Exception: ' + e);
		}
		
		return true;
	},
	
	prepareUsagePageOpen: function() {
		var start = 0, usagepage = -1, lastgp = 0, nextisusagepage = false;
		for (var i = 0; i < this.isp.process.length; i++) {
			if (typeof(this.isp.process[i]) == 'string') {
				if (this.isp.process[i] == 'STARTUSAGEPAGE')
					start = i+1;
				else if (this.isp.process[i] == 'USAGEPAGE')
					nextisusagepage = true;
				else if (this.isp.process[i].substring(0, 3) == 'GET' || // GETBASIC POSTBASIC
					this.isp.process[i].substring(0, 4) == 'POST') {
					if (nextisusagepage)
						usagepage = i;
					else
						lastgp = i;
					nextisusagepage = false;
				}
			}
		}
		ndump("lastgp "+lastgp+" usagepage "+usagepage+" start "+start);
		if (usagepage == -1)
			usagepage = lastgp;
		this.mUsagePageStep = usagepage;
		this.step = start;
		this.nextstep = start;
	},
	
	resumeProcess: function() {
		var gonext = true; // needed because asynchronous XMLHttpRequest breaks out of process loop
		while (gonext) {
			this.vars['laststep'] = this.step;
			this.step = this.nextstep;
			
			gonext = this.stepProcess();
			
			if (this.mStatus != 'fetching') {
				// process step error occured OR all done
				this.endUpdate(null, null);
				break;
			}
			ndump("resumeProcess done step "+this.step);
		}
		ndump("resumeProcess end");
	},

	// steps this.isp.process returns true if caller should immediately call next
	stepProcess: function() {
		ndump("stepProcess() step "+this.step);
		
		// check loop LOOP MAX = 5
		if (typeof(this.vars['step' + this.step + 'count']) == 'undefined') {
			this.vars['step' + this.step + 'count'] = 0;
		}
		this.vars['step' + this.step + 'count']++;
		
		if (this.vars['step' + this.step + 'count'] > 5) {
			ndump("stepUpdateUsage() step loop maximum exceeded");
			
			this.mStatus = 'error-process';
			this.mErrorText = 'Step looping maximum exceeded!';
			return false;
		}
		
		if (typeof(this.step) == 'string') {
			// resolve label
			for (var i in this.isp.process) {
				if (typeof(this.isp.process[i]) == 'string' && this.isp.process[i] == "LABEL "+this.step) {
					this.step = parseInt(i)+1;
					break;
				}
			}
		}
		if (typeof(this.step) == 'string') { // still eq string, then label not found
			this.mStatus = 'error-process';
			this.mErrorText = 'Unable to find LABEL '+this.step;
			return false;
		}
		
		this.nextstep = this.step + 1;

		// NOW See what step we are up to and what type of step is it
		if (this.step >= this.isp.process.length) {
			ndump("process step finish");
			// we've arrived at the finish
			// post process
			if (!this.mVarMapped)
				this.mapVariables();
			
			ndump("map var finish");
			// update last fetched
			this.mLastFetched = Math.floor((new Date()).getTime()/1000);
			
			// debug
			ndump('this.vars = {');
			for (i in this.vars) {
				if (i == 'USERNAME' || i == 'PASSWORD')
					continue;
				var zzz = this.vars[i];
				if (typeof(zzz) == 'string') {
					zzz = zzz.substr(0, 100);
				}
				ndump(i+': '+zzz);
			}
			ndump('}');

			this.mStatus = 'done';
			return false;
		}
		// process step function
		else if (typeof(this.isp.process[this.step]) == 'function' && typeof(this.isp.process[this.step]['exec']) != 'function') {
			ndump("process step function");
			
			var ns = (this.isp.process[this.step])(this.vars);
			if (typeof(ns) == 'number' || typeof(ns) == 'string')
				this.nextstep = ns;

			if (this.vars['errortext']) {
				// process function raised an error
				this.mStatus = 'error-isp';
				this.mErrorText = this.vars['errortext'];
				return false;
			}
		}
		// process step command
		else if (typeof(this.isp.process[this.step]) == 'string') {
			ndump("process step command");
			// substitute all vars in URL cmd string
			var s = nuVarExpand(this.isp.process[this.step], this.vars, true, false, false, this.isp, 
				this.filterFuncList, this.filterFunc);
			
			if (this.mUsagePageStep == this.step) {
				// we arrived at the usage page step, open usage page in GUI
				this.mOpenPageCmd = s;
				this.nextstep = 1000;
				this.mStatus = 'done';
				return false;
			}

			var params = s.split(/ /);
			if (params[0] == 'POST' || params[0] == 'GET' || params[0] == 'POSTBASIC' || params[0] == 'GETBASIC') {
				this.stepUpdateUsage(params);
				return false;
			}
			else if (params[0] == 'MAPVARS')
				this.mapVariables();
			else if (params[0] == 'LABEL')
				{} // do nothing, skip
			else if (params[0] == 'STARTUSAGEPAGE') {
				// all the steps below this is usage page fetching related
				// here we are not fetching usage page, don't proceed
				this.nextstep = 1000;
			}
			else if (params[0] == 'USAGEPAGE') {
				// here, the next action should be notifying the GUI to show the usage page
				// all logic is at prepareUsagePageOpen though
			}
			else if (params[0] == 'REFRESHREDIRECT') {
				// handle http equiv redirects
				var ra = / content=(?:["']|)[0-9]+;url=(?:["']|)([^"']+)/i.exec(this.xmlhttp.responseText);
				if (ra) {
					this.stepUpdateUsage(['GET', ra[1]]);
					this.nextstep = this.step;
					return false;
				}
			}
			else {
				ndump('UKNOWN COMMAND '+params[0]);
				this.mStatus = 'error-process';
				this.mErrorText = 'Unknown process command (S'+this.step+'): '+s;
				return false;
			}
		}
		// process step parsing
		else if (typeof(this.isp.process[this.step]) == 'object' || typeof(this.isp.process[this.step]['exec']) == 'function') {
			ndump("process step parsing");
			this.nextstep = this.parse(this.isp.process[this.step]);

			if (this.nextstep == 'ERROR') {
				if (this.mStatus != 'error-parse') {
					if (this.vars[this.getParseNum(this.step) + ":1"]) {
						this.mErrorText = 'ISP Error: '+this.vars[this.getParseNum(this.step) + ":1"];
					} else {
						this.mErrorText = 'Login Failed. Please check your username and password.';
					}
					this.mStatus = 'error-isp';
				}
				this.nextstep = -1;
				return false;
			}
		}
		
		ndump('nextstep: '+this.nextstep);
		return true;
		
	},
		
	stepUpdateUsage: function(params) {
		var poststr;

		if (params[0] == 'POST' || params[0] == 'POSTBASIC')
			poststr = params[2];
		else
			poststr = null;

		this.xmlhttp = Components.classes['@mozilla.org/xmlextras/xmlhttprequest;1'].createInstance(Components.interfaces.nsIXMLHttpRequest);

		//this.xmlhttp.QueryInterface(Components.interfaces.nsIDOMEventTarget);
		this.xmlhttp.addEventListener('load', this, false);
		this.xmlhttp.addEventListener('error', this, false);
		
		ndump('OPEN '+params[0]+';'+params[1]);
		this.xmlhttp.open(params[0] == 'GETBASIC' ? 'GET' : (params[0] == 'POSTBASIC' ? 'POST' : params[0]) , params[1]);
		
		// throw off caching
		this.xmlhttp.setRequestHeader("If-Modified-Since", "Sat, 1 Jan 2000 00:00:00 GMT");
		//this.xmlhttp.setRequestHeader("Referer", ); TODO
		
		if (params[0] == 'POST' || params[0] == 'POSTBASIC')
			this.xmlhttp.setRequestHeader('Content-type', 'application/x-www-form-urlencoded');
		if (params[0] == 'GETBASIC' || params[0] == 'POSTBASIC')
			this.xmlhttp.setRequestHeader('Authorization', 'Basic ' + nuEncodeBase64(this.vars['USERNAME']+':'+this.vars['PASSWORD']));

		this.mDebughttp += this.step+') '+params[0]+' '+params[1]+' '+(poststr != null ? poststr : '')+"\n";
		this.xmlhttp.send(poststr);

	},
	
	debugDump: function(strprefix) {
		// dump response, source into profile
		var doc = strprefix+"Version "+nuGetVersion()+"\n"+this.mDebughttp+"\n\n"+ndumpobjstr('this.vars', this.vars)+"\nNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNN\n";
		if (this.xmlhttp)
			doc += this.xmlhttp.responseText;
		
		// sanitize transcript to protect u/p
		if (typeof(this.vars['USERNAME']) != 'undefined' && this.vars['USERNAME'] != null && this.vars['USERNAME'] != '') {
			doc = doc.replace(new RegExp(nuEscapeRegexp(encodeURIComponent(this.vars['USERNAME'])), 'g'), 'UXXX');
			doc = doc.replace(new RegExp(nuEscapeRegexp(this.vars['USERNAME']), 'g'), 'UXXX');
		} 
		if (typeof(this.vars['PASSWORD']) != 'undefined' && this.vars['PASSWORD'] != null && this.vars['PASSWORD'] != '') {
			doc = doc.replace(new RegExp(nuEscapeRegexp(encodeURIComponent(this.vars['PASSWORD'])), 'g'), 'PXXX');
			doc = doc.replace(new RegExp(nuEscapeRegexp(this.vars['PASSWORD']), 'g'), 'PXXX');
		}
		
		var file = Components.classes["@mozilla.org/file/directory_service;1"].getService(Components.interfaces.nsIProperties).get("ProfD", Components.interfaces.nsIFile);
		file.append("netusage_debug.txt");
		var fs = Components.classes["@mozilla.org/network/file-output-stream;1"].createInstance(Components.interfaces.nsIFileOutputStream);
		fs.init(file, 0x02 | 0x08 | 0x20, 0664, 0); // write, create, truncate
		fs.write(doc, doc.length);
		fs.close();
	},
	
	getParseNum: function(step) {
		var parsenum = 0;
		for (var i = 0; i < step; i++) {
			if (typeof(this.isp.process[i]) == 'object' || typeof(this.isp.process[i]['exec']) == 'function')
				parsenum++;
		}
		return parsenum;
	},

	// parsing, called by dom event load
	parse: function(ae) {
		var rv, i;
		var branch = 0;
		var parsenum = this.getParseNum(this.step);
		if (typeof(ae) == 'object') {
			ndump("ae typeof object/array");
			
			// 3rd arg is an array... loop through it
			for (i in ae) {
				// theoretically, we can mix xpath expr and regexp expr leafs
				if (typeof(ae[i]['exec']) == 'function') {
					// one regexp out of multiple (regexp, xpath)
					ndump("one regexp out of multiple");
					rv = this.parseRegexp(ae[i], this.xmlhttp.responseText);
				}
				else if (typeof(ae[i]) == 'object') {
					// one xpath out of multiple (regexp, xpath)
					ndump("one xpath out of multiple");
					rv = this.parseXML(ae[i], this.xmlhttp.responseXML);
				}
				else if (typeof(ae[i]) == 'string' && ae["XPATHVERIFY"]) {
					// only 1 xpath branch, (one xpath branch is a string array of xp expr)
					ndump("1 xpath branch");
					rv = this.parseXML(ae, this.xmlhttp.responseXML);
					this.vars['parse'+parsenum] = 0; // the branch we took
					break;
				}
				else if (typeof(ae[i]) == 'string' || typeof(ae[i]) == 'number') {
					// this is a special directive of the path to take if we match
					branch--;
				}
				
				// did we find a match?
				if (rv) {
					this.vars['parse'+parsenum] = branch; // the branch we took
					break;
				}
				branch++;
			}
		}
		else if (typeof(ae['exec']) == 'function') {
			// only 1 regexp
			ndump("ae[exec] typeof regexp");
			rv = this.parseRegexp(ae, this.xmlhttp.responseText);
			this.vars['parse'+parsenum] = 0;
		}

		if (!rv) {
			// parsing failed
			ndump("parsing failed");
			this.mStatus = 'error-parse';
			this.mErrorText = 'A parsing error occured (S'+this.step+')';
			return 'ERROR';
		}
		
		// move stuff in rv into this.vars, append 'parsenum:' to key
		for(var j in rv) {
			this.vars[parsenum  + ":" + j] = rv[j];
		}
		
		if (ae["XPATHVERIFY"] != undefined)
			return this.nextstep;
		var ni = parseInt(i) + 1;
		return (typeof(ae[ni]) == 'number' || typeof(ae[ni]) == 'string') ? ae[ni] : this.nextstep;
	},


	// parsing text using regexp
	parseRegexp: function(aExp, aDoc) {
		var a = aExp.exec(aDoc);
		if (!a) {
			ndump('parseRegexp failed');
			return null;
		}
		a[0] = 0;
		return a;
	},


	// parsing XML using xpath
	parseXML: function(aPaths, aDoc) {
		try {
			if (aPaths['XPATHVERIFY'] && !aDoc.evaluate(aPaths['XPATHVERIFY'], aDoc, null, 9, null).singleNodeValue) {
				ndump('parse XPATH failed');
				return null;
			}
		} catch(e) {
			ndump('parse XPATH failed ' + e);
			return null;
		}
		for (var k in aPaths) {
			if (k == 'XPATHVERIFY') continue;
			nuVarPipe(k, aDoc.evaluate(aPaths[k], aDoc, null, 2, null).stringValue, this.vars, false, false, true, this.isp, 
				this.filterFuncList, this.filterFunc);
		}
		return [];
	},
	
	// get the start and end dates as an array [start, end]
	getStartEndDate: function(vstart, vend, vroll) {
		var startdate, enddate;
		var now = new Date();

		if (vroll) {
			// only roll date AKA next start date specified
			rolldate = this.isp.strToDate('roll', vroll, this.isp);
			enddate = new Date(rolldate.getTime()-1);
			startdate = new Date(enddate.getFullYear(), enddate.getMonth(), rolldate.getDate(), 0, 0, 0);
			if (startdate.getTime() > enddate.getTime())
				startdate = new Date(enddate.getFullYear(), enddate.getMonth()-1, rolldate.getDate(), 0, 0, 0);
		}
		else if (!vstart && vend) {
			// only end date specified
			enddate = this.isp.strToDate('end', vend, this.isp);
			startdate = new Date(enddate.getFullYear(), enddate.getMonth()-1, enddate.getDate()+1, 0, 0, 0);
		}
		else if (vstart && !vend) {
			// only start date specified
			startdate = this.isp.strToDate('start', vstart, this.isp);
			enddate = new Date(startdate.getFullYear(), startdate.getMonth()+1, startdate.getDate()-1, 23, 59, 59);
		}
		else if (vstart && vend) {
			// both start and end date specified
			startdate = this.isp.strToDate('start', vstart, this.isp);
			enddate = this.isp.strToDate('end', vend, this.isp);
		}
		else {
			// calendar month
			startdate = new Date(now.getFullYear(), now.getMonth(), 1, 0, 0, 0);
			enddate = new Date(now.getFullYear(), now.getMonth()+1, 0, 23, 59, 59);
		}
		return [startdate, enddate];
	},
	
	// map these variables
	mapVariables: function(vstart, vend, vroll) {
		this.mVarMapped = true;
		
		var a;
		// var mapping
		for (var i = 0; i < this.isp.process.length; i++) {
			if (a = this.isp.varsDisplay[i+','+this.vars["parse"+i]])
				for(k in a) {
					if (typeof(a[k]) == 'object') {
						nuVarExpand(a[k][0], this.vars, false, false, true, this.isp, 
							this.filterFuncList, this.filterFunc);
						nuVarExpand(a[k][1], this.vars, false, false, true, this.isp, 
							this.filterFuncList, this.filterFunc);
					} else
						nuVarExpand(a[k], this.vars, false, false, true, this.isp, 
							this.filterFuncList, this.filterFunc);
				}
		}
		
		// usage var doesn't exist, default to download var
		if (typeof(this.vars['usage-quota']) == 'undefined') {
			this.vars['usage'] = this.vars['download'];
			this.vars['usage-quota'] = this.vars['download-quota'];
		}
		if (typeof(this.vars['usagepo-quota']) == 'undefined' && typeof(this.vars['downloadpo-quota']) != 'undefined') {
			this.vars['usagepo'] = this.vars['downloadpo'];
			this.vars['usagepo-quota'] = this.vars['downloadpo-quota'];
		}

		// start end date inference
		var d = this.getStartEndDate(this.vars['startdate'], this.vars['enddate'], this.vars['rolldate']);

		this.vars['startdate'] = Math.floor(d[0].getTime()/1000);
		this.vars['enddate'] = Math.floor(d[1].getTime()/1000);
		
		// adjust to ISP timezone
		if (typeof(this.vars['isptz']) != 'undefined' && this.vars['isptz'] != 86400) {
			this.vars['startdate'] -= d[0].getTimezoneOffset()*60 + this.vars['isptz'];
			this.vars['enddate'] -= d[0].getTimezoneOffset()*60 + this.vars['isptz'];
		}
	},
	
	filterFuncList: ['B', 'KB', 'MB', 'GB', 'MiB', 'COMMADEC'],
	filterFunc: function(f, s) {
		if (!s) return 0;
		ndump("filterfunc START "+f+" "+s);
		if (typeof(s) == 'number') s = s+"";
		ndump("filterfunc END");
		
		// convert euro comma decimal points to dot decimal point
		if (f == 'COMMADEC')
			return s.replace(/,/g, '.');
		
		// axe away commas
		s = s.replace(/,/g, '');
		// we're using MB as std unit because getIntPref is 32bit only
		if (f == 'B')
			return Math.round(parseInt(s, 10)/1000000);
		else if (f == 'KB')
			return Math.round(parseInt(s, 10)/1000);
		else if (f == 'GB')
			return Math.round(parseFloat(s)*1000);
		else if (f == 'MB') // assume mbytes
			return parseInt(s, 10);
		else if (f == 'MiB') // assume mbytes
			return Math.round(parseInt(s, 10)*1024*1024/1000000);
	},


	// load user preferences
	loadPrefs: function () {
		var doinittime = false;

		try {
			var prefSvc = Components.classes["@mozilla.org/preferences-service;1"].getService(Components.interfaces.nsIPrefService);
			var pref = prefSvc.getBranch("extensions.netusage."+this.mitemid+".");
			
			try {
				this.mDebug = pref.getBoolPref("debug");
			} catch(ee) { }
			
			var prefISP = pref.getCharPref("userisp");
			if (prefISP && prefISP != '' && prefISP != 'na') {
				if (this.isp == null || this.isp.id != prefISP)
					this.isp = this.getISPInterface(prefISP);

			} else {
				if (this.isp != null) {
					this.isp = null;
					doinittime = true; // disable the timer
				}
			}
			var prefCheckInterval = pref.getIntPref("checkinterval");
			// if timer checkInterval is less than 10 mins (minimum req, so ISP servers aren't DOS'd)
			// initTimer() will disable the timer
			if (this.checkInterval != prefCheckInterval) {
				this.checkInterval = prefCheckInterval;
				doinittime = true; // refresh timer interval
			}
			if (doinittime)
				this.initTimer();

		} catch(e) { ndump('SOFTFAIL ' + e); }


	},


	//
	// nsIObserver
	//
	observe: function (subject, topic, data) {
		ndump('Observed: '+subject+' :: '+topic+' ::: '+data);
		switch (topic) {
			case 'timer-callback':
				this.updateUsage();
				break;
			case 'nsPref:changed':
				this.loadPrefs();
				break;
			case 'http-on-examine-response':
				subject.QueryInterface(Components.interfaces.nsIHttpChannel);
				// if our channel
				if (this.xmlhttp.channel == subject) {
					// nuke WWW-Authenticate if exist so ugly dialog doesn't pop up if wrong pass
					try {
					if (subject.getResponseHeader('WWW-Authenticate')) {
						subject.setResponseHeader('WWW-Authenticate', '', false);
						ndump('WWW-Authenticate safely contained');
					}
					} catch(e) { ndump('SOFTFAIL: '+e); }
					
					var hcontenttype, hlocation;
					try {
					hcontenttype = subject.getResponseHeader("Content-Type");
					} catch(e) { hcontenttype = '';}
					try {
					hlocation = ' -> '+subject.getResponseHeader("Location");
					} catch(e) { hlocation = ''; }
					
					this.mDebughttp += this.step+') '+hcontenttype+' '+subject.responseStatus+' '+subject.URI.spec+hlocation+"\n";
				}
				break;
		}
	},



	//
	// nsIDOMEventListener
	//
	handleEvent: function(event) {
		ndump("handleEvent: "+event.type);
		
		// check if we aborted
		if (this.mStatus == 'aborted') {
			//exit
			// TODO WE NEVER REACH HERE?
			ndump('aborted');
			//this.observerSvc.notifyObservers(this, 'netusage-update-gui', 'draw');
			return;
		}
		
		try {

		if (event.type == 'load') {
			// XMLHttpLoadEvent
			if (this.xmlhttp.status != 200 && this.xmlhttp.status != 0 /* for file:/// */) {
				// HTTP error occured
				this.mStatus = 'error-isp';
				if (this.xmlhttp.status == 401) {
					// authorization required
					this.mErrorText = 'Incorrect username or password.';
				} else {
					this.mErrorText = 'ISP error occured: '+this.xmlhttp.status+' '+this.xmlhttp.statusText;
				}
				this.endUpdate(null, null);
			}
			else
				this.resumeProcess();
		}
		else if (event.type == 'error') {
			// XMLHttpErrorEvent
			this.endUpdate('error', 'Connection error occured (S'+this.step+')');
		}
		
		} catch(e) {
			this.endUpdate('error', 'Exception: '+e+' (S'+this.step+')');
		}

	},


	


	//
	// nsISupports
	//


	QueryInterface: function (iid) {
		if (!iid.equals(Components.interfaces.nsINetUsageItem) &&
			!iid.equals(Components.interfaces.nsISupports) &&
			!iid.equals(Components.interfaces.nsIObserver) &&
			!iid.equals(Components.interfaces.nsIDOMEventListener)) {
			throw Components.results.NS_ERROR_NO_INTERFACE;
		}
		return this;
	},
}



function nsNetUsageItem() {

	this.xmlhttp = null;
	this.timer = null;

	this.step = 0;
	this.vars = null;
	this.mLastFetched = 0;
	this.mStatus = "";
	this.mDebug = false;



}





// MODULE
var module = {
	firstTime: true,

	cid: Components.ID(nui_class_id),

	contractId: nui_contract_id,

	registerSelf: function (compMgr, fileSpec, location, type) {
		if (this.firstTime) {
			// defer first time registering, I have no idea why though :/
			this.firstTime = false;
			throw Components.results.NS_ERROR_FACTORY_REGISTER_AGAIN;
		}
		compMgr = compMgr.QueryInterface(Components.interfaces.nsIComponentRegistrar);
		compMgr.registerFactoryLocation(this.cid, "Net Usage Component", this.contractId, fileSpec, location, type);
	},

	unregisterSelf: function(compMgr, fileSpec, location) {
		compMgr = compMgr.QueryInterface(Components.interfaces.nsIComponentRegistrar);
		compMgr.unregisterFactoryLocation(this.cid, fileSpec);
	},

	canUnload: function(compMgr) {
		return true;
	},

	// getClassObject: returns component's factory object
	getClassObject: function (compMgr, cid, iid) {
		if (!cid.equals(this.cid)) {
			throw Components.results.NS_ERROR_NO_INTERFACE;
		}

		if (!iid.equals(Components.interfaces.nsIFactory)) {
			throw Components.results.NS_ERROR_NOT_IMPLEMENTED;
		}

		return this.factory;
	},

	// factory object
	factory: {
		createInstance: function (outer, iid) {
			if (outer != null)
				throw Components.results.NS_ERROR_NO_AGGREGATION;
			dump('>>>> NUI CREATED\n');
			return (new nsNetUsageItem()).QueryInterface(iid);
		}
	}
};

function NSGetModule(compMgr, fileSpec) {
	return module;
}
